#Inserting information in the fields below is optional
$fullPathToFiles = @(); #e.g. @("$/Designs/example.iam", ...)

$userName = "" 
$password = ""              
$serverName = ""
$vaultName = ""

$outputPath = ""

function OpenVaultConnectionViaGUI() {
  $settings = [Autodesk.DataManagement.Client.Framework.Vault.Forms.Settings.LoginSettings]::new()
  $VaultConnection = [Autodesk.DataManagement.Client.Framework.Vault.Forms.Library]::Login($settings)
  if($VaultConnection -eq $null) {
    Write-Error "Connecting to Vault was canceled!"
    return
  }
  return $VaultConnection
}

function SelectVaultPackageViaGUI() {
  $settings =  [Autodesk.DataManagement.Client.Framework.Vault.Forms.Settings.SelectEntitySettings]::new()
  $files = [Autodesk.DataManagement.Client.Framework.Vault.Forms.Library]::SelectEntity($vaultConnection, $settings)
  if($files -eq $null) {
    Write-Error "Vault file selection was canceled!"
    return
  }
  return $files
}

function SelectOutputPathViaGUI() {
  [System.Reflection.Assembly]::LoadWithPartialName("System.windows.forms") > $null
  $folderBrowser = New-Object System.Windows.Forms.FolderBrowserDialog
  $folderBrowser.Description = "Select the output folder"
  $folderBrowser.RootFolder = [Environment+SpecialFolder]::Desktop
  $folderBrowser.ShowDialog() | Out-Null
  if ($folderBrowser.SelectedPath -ne $null -and [string]::IsNullOrWhiteSpace($folderBrowser.SelectedPath) -eq $false -and [System.IO.Directory]::Exists($folderBrowser.SelectedPath)) {
     return $folderBrowser.SelectedPath
  } else {
     Write-Error "Invalid output path specified!"
     return
  }
}

function WriteFileToJson($file, $wsm) {
  if ($fileIds.Contains($file.Id)){
    return;
  }

  $path = Join-Path $TEMP_PATH $file.Name;
  [System.IO.Directory]::CreateDirectory($path) | Out-Null

  $fullpath = Join-Path $path "$($file.Id).json";
  $serializer = New-Object -TypeName System.Web.Script.Serialization.JavaScriptSerializer
  $json = $serializer.Serialize($file);
  Set-Content $fullpath $json -NoNewline

  $bom = $wsm.DocumentService.GetBOMByFileId($file.Id);
  $serializer = New-Object -TypeName System.Web.Script.Serialization.JavaScriptSerializer
  $json = $serializer.Serialize($bom);
  $fullpath = Join-Path $path "$($file.Id)_bom.json";
  Set-Content $fullpath $json -NoNewline

  $assocsFileIds = [System.Collections.ArrayList]@();
  $assocsFileIds.Add($file.Id) | Out-Null

  $assocs = $wsm.DocumentService.GetFileAssociationsByIds($assocsFileIds, [Autodesk.Connectivity.WebServices.FileAssociationTypeEnum]::None, $false, [Autodesk.Connectivity.WebServices.FileAssociationTypeEnum]::Dependency, $true, $false, $false);
  if ($assocs -ne $null -and $assocs[0] -ne $null)
  {
    $serializer = New-Object -TypeName System.Web.Script.Serialization.JavaScriptSerializer
    $json = $serializer.Serialize($assocs[0]);
    $fullpath = Join-Path $path "$($file.Id)_assoc.json";
    Set-Content $fullpath $json -NoNewline
  }

  $path = Join-Path $TEMP_PATH $file.Name;
  $fullpath = Join-Path $path "$($file.Id)_vp.json";
  $fullName = GetVaultFolderPathByMasterId $file.MasterId $wsm;
  Set-Content $fullpath $fullname -NoNewline

  $fileIds.Add($file.Id) | Out-Null
}

function RecurseFiles($fileMasterId, $vault) {
  if ($fileMasterIds.Contains($fileMasterId)){
    return;
  }
  $fileMasterIds.Add($fileMasterId) | Out-Null

  $files = $vault.DocumentService.GetFilesByMasterId($fileMasterId);
  foreach ($file in $files){
      WriteFileToJson $file $vault;
  }

  $assocsFileIds = @($files | select -Property id | foreach {$_.Id})
  $assocs = $vault.DocumentService.GetFileAssociationsByIds($assocsFileIds, [Autodesk.Connectivity.WebServices.FileAssociationTypeEnum]::None, $false, [Autodesk.Connectivity.WebServices.FileAssociationTypeEnum]::Dependency, $false, $false, $false);
  foreach ($assoc in $assocs) {
      if ($assoc.FileAssocs -ne $null) {
         foreach ($fileAssoc in $assoc.FileAssocs) {
             RecurseFiles $fileAssoc.CldFile.MasterId $vault;
         }
      }
  }

  $boms = [System.Collections.ArrayList]@();	
  foreach($file in $files) {
   	 $bom = $vault.DocumentService.GetBOMByFileId($file.Id);
		 if($bom -eq $null) {
		   Write-Warning "[WARN] - File '$($file.Name)' with Id '$($file.Id)' has no BOM!";
			 continue;
		 }
		 $boms.Add($bom) | Out-Null
  }

  foreach ($bom in $boms) {
    $bomCompArrayWithoutInternal = $boms.CompArray | select -Unique -ExpandProperty XRefId | Where-Object {$_ -ne -1}
    foreach ($fileId in $bomCompArrayWithoutInternal) {
          try {
            $file = $vault.DocumentService.GetFileById($fileId);
            RecurseFiles $file.MasterId $vault;
          }
          catch {
          }
    }
  }
}

function packOutputToZip($outputPath, $filename) {
  if (Test-Path $TEMP_PATH)
  {
    $zipPath = Join-Path $outputPath "$($filename)-bomExtract.zip";
    Compress-Archive -Path $TEMP_PATH -CompressionLevel Fastest -DestinationPath $zipPath
  }
}

function cleanUp {
  if (Test-Path $TEMP_PATH)
  {
    Remove-Item -Recurse -Force $TEMP_PATH;
  }
}

function GetVaultFolderPathByMasterId($fileMasterId, $vault) {
  $folderId = $vault.DocumentService.GetLatestFileByMasterId($fileMasterId);
  $folderId = $folderId.FolderId;
  $fullName = $vault.DocumentService.GetFolderById($folderId);
  $fullName = $fullName.FullName;
  return $fullName
}

function ExtractBomForFiles($fileMasterId, $vault, $outputPath) {
  RecurseFiles $fileMasterId $vault;
  $filename = $vault.DocumentService.GetLatestFileByMasterId($fileMasterId);
  $filename = $filename.Name;
  $pathRootName = Join-Path $TEMP_PATH "rootFile.txt";
  $pathRootVault = Join-Path $TEMP_PATH "rootVaultPath.txt";
  Set-Content $pathRootName $filename -NoNewline;
  $fullName = GetVaultFolderPathByMasterId $fileMasterId $vault;
  Set-Content $pathRootVault $fullname -NoNewline;
  packOutputToZip $outputPath $filename;
  cleanUp;
  Write-Host "Successfully exported BOMs of file: '$filename'"
}

$TEMP_PATH = "C:\BOMExtractor\"
$fileIds = [System.Collections.Generic.List[long]]::new();
$fileMasterIds = [System.Collections.Generic.List[long]]::new();

Import-Module powerVault
Add-Type -AssemblyName "System.Web.Extensions"
Add-Type -AssemblyName "Autodesk.DataManagement.Client.Framework.Vault.Forms"

if ([string]::IsNullOrEmpty($vaultName) -eq $true -and 
[string]::IsNullOrEmpty($serverName) -eq $true -and 
[string]::IsNullOrEmpty($userName) -eq $true -and 
[string]::IsNullOrEmpty($password) -eq $true) {
  $VaultConnection = OpenVaultConnectionViaGUI
  if($VaultConnection -eq $null) {
    return
  }
  Open-VaultConnection
} else {
  Open-VaultConnection -Vault $vaultName -Server $serverName -User $userName -Password $password
}

if ($fullPathToFiles.Count -eq 0) {
  $selectedEntities = SelectVaultPackageViaGUI $vaultConnection
  if ($selectedEntities -eq $null) {
    return
  }
  $vaultFolderPath = $(GetVaultFolderPathByMasterId $selectedEntities.SelectedEntities.EntityMasterId $vault)
  $fullPathToFiles = "$($vaultFolderPath)/$($selectedEntities.SelectedEntities.EntityName)"
}

foreach ($fullPathToFile in $fullPathToFiles) {
  $files = Get-VaultFile -File $fullPathToFile
  if($files.Count -eq 0 -or $files.Count -gt 1) {
    Write-Error "No or multiple Vault files with the same name found!"
    return
  }


  if ([string]::IsNullOrEmpty($outputPath) -eq $true) {
    $outputPath = SelectOutputPathViaGUI
    if([string]::IsNullOrEmpty($outputPath) -eq $true) {
      return
    }
  }

  # Create root target dir
  [System.IO.Directory]::CreateDirectory($TEMP_PATH) > $null
  ExtractBomForFiles $files[0].MasterId $vault $outputPath

}